#include "AES_SIMD.h"
#include <memory.h>

//Cipher(byte in[4*Nb], byte out[4*Nb], word w[Nb*(Nr+1)])
//begin
void AES_SIMD::Cipher(const BYTE in[4 * Nb], BYTE out[4 * Nb])
{
//	ASSERT(!decrypting);

	//state = in
	__m128i state = _mm_loadu_si128(reinterpret_cast<const __m128i *>(in));

	//AddRoundKey(state, w[0, Nb-1])
	// xorZ邾
	state = _mm_xor_si128(state, w.w128[0]);

	//for round = 1 step 1 to Nr-1
	for (int round = 1; round <= Nr - 1; round++) {
		//SubBytes(state)
		//ShiftRows(state)
		//MixColumns(state)
		//AddRoundKey(state, w[round*Nb, (round+1)*Nb-1])
		state = _mm_aesenc_si128(state, w.w128[round]);

	//end for
	}

	// Ō̃Eh
	//SubBytes(state)
	//ShiftRows(state)
	//AddRoundKey(state, w[Nr*Nb, (Nr+1)*Nb-1])
	state = _mm_aesenclast_si128(state, w.w128[Nr]);

	//out = state
	_mm_storeu_si128(reinterpret_cast<__m128i *>(out), state);

//end
}

//EqInvCipher(byte in[4*Nb], byte out[4*Nb], word dw[Nb*(Nr+1)])
//begin
void AES_SIMD::EqInvCipher(const BYTE in[4 * Nb], BYTE out[4 * Nb])
{
//	ASSERT(decrypting);

	//byte  state[4,Nb]
	//state = in
	__m128i state = _mm_loadu_si128(reinterpret_cast<const __m128i *>(in));

	//AddRoundKey(state, dw[Nr*Nb, (Nr+1)*Nb-1])
	state = _mm_xor_si128(state, dw.w128[Nr]);

	//for round = Nr-1 step -1 downto 1
	for (int round = Nr - 1; round >= 1; round--) {
		//InvSubBytes(state)
		//InvShiftRows(state)
		//InvMixColumns(state)
		//AddRoundKey(state, dw[round*Nb, (round+1)*Nb-1])
		state = _mm_aesdec_si128(state, dw.w128[round]);

	//end for
	}

	//InvSubBytes(state)
	//InvShiftRows(state)
	//AddRoundKey(state, dw[0, Nb-1])
	state = _mm_aesdeclast_si128(state, dw.w128[0]);

	//out = state
	_mm_storeu_si128(reinterpret_cast<__m128i *>(out), state);
//end
}

//static const BYTE Rcon[] = {
//	0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a,
//};

//KeyExpansion(byte key[4*Nk], word w[Nb*(Nr+1)], Nk)
//begin
void AES_SIMD::KeyExpansion(const BYTE key[4 * Nk])
{
	//word  temp
	//i = 0

	//while (i < Nk)
	//w[i] = word(key[4*i], key[4*i+1], key[4*i+2], key[4*i+3])
	//i = i+1
	//end while
	memcpy(w.w32, key, 4 * Nk);

	//i = Nk
	int i = Nk;
	//while (i < Nb * (Nr+1)]
	for (; i < Nb * (Nr + 1); i++) {
		//temp = w[i-1]
		AESWORD temp = w.w32[i - 1];

		//if (i mod Nk = 0)
		if (i % Nk == 0) {
			//temp = SubWord(RotWord(temp)) xor Rcon[i/Nk]
			__m128i t;
			t.m128i_u32[1] = temp;
			switch (i / Nk) {
			case 1:
				t = _mm_aeskeygenassist_si128(t, 0x01);
				break;
			case 2:
				t = _mm_aeskeygenassist_si128(t, 0x02);
				break;
			case 3:
				t = _mm_aeskeygenassist_si128(t, 0x04);
				break;
			case 4:
				t = _mm_aeskeygenassist_si128(t, 0x08);
				break;
			case 5:
				t = _mm_aeskeygenassist_si128(t, 0x10);
				break;
			case 6:
				t = _mm_aeskeygenassist_si128(t, 0x20);
				break;
			case 7:
				t = _mm_aeskeygenassist_si128(t, 0x40);
				break;
			case 8:
				t = _mm_aeskeygenassist_si128(t, 0x80);
				break;
			case 9:
				t = _mm_aeskeygenassist_si128(t, 0x1b);
				break;
			case 10:
				t = _mm_aeskeygenassist_si128(t, 0x36);
				break;
			default:
//				ASSERT(0);
				break;
			}

			temp = t.m128i_u32[1];
		}
		//else if (Nk > 6 and i mod Nk = 4)
		else if (Nk > 6 && (i % Nk) == 4) {
			//temp = SubWord(temp)
			__m128i t;
			t.m128i_u32[1] = temp;
			t = _mm_aeskeygenassist_si128(t, 0);
			temp = t.m128i_u32[0];
		}
		//end if

		//w[i] = w[i-Nk] xor temp
		w.w32[i] = w.w32[i - Nk] ^ temp;

		//i = i + 1
	//end while
	}

	// EqInvCipherp̒ǉ
	if (decrypting) {
		//for i = 0 step 1 to (Nr+1)*Nb-1
			//dw[i] = w[i]
		//end for
		dw = w;

		//for round = 1 step 1 to Nr-1
		for (int round = 1; round <= Nr - 1; round++) {
			//InvMixColumns(dw[round*Nb, (round+1)*Nb-1])
			dw.w128[round] = _mm_aesimc_si128(dw.w128[round]);
		//end for
		}
	}
//end
}

