#include <iostream>
#include <string>
#include <vector>
#include <bitset>
#include "AES_SIMD.h"
#include "AES128_SIMD.h"
#include "AES192_SIMD.h"
#include "AES256_SIMD.h"

using namespace std;

typedef unsigned char BYTE;

vector<BYTE> ByteArrayFromString(const string& str)
{
	vector<BYTE> arr;

	static const BYTE INVALID_CHAR = 0xFF;

	auto HexVal = [](char c) -> BYTE {
		if (c >= '0' && c <= '9') return c - '0';
		else if (c >= 'a' && c <= 'f') return c - 'a' + 10;
		else if (c >= 'A' && c <= 'F') return c - 'A' + 10;
		else return INVALID_CHAR;
	};

	for (size_t i = 0; i + 1 < str.length(); i += 2) {
		BYTE v1 = HexVal(str[i]);
		BYTE v2 = HexVal(str[i + 1]);
		if (v1 == INVALID_CHAR || v2 == INVALID_CHAR)
			break;
		arr.push_back(v1 << 4 | v2);
	}
	return arr;
}


template<class AES_CLASS>
void test()
{
	// FIPS 197 t^C.1̃eXgf[^gp

	string key, plain, cipher;

	if (AES_CLASS::Nk == 4) {
		// AES128
		key = "000102030405060708090a0b0c0d0e0f";
		plain = "00112233445566778899aabbccddeeff";
		cipher = "69c4e0d86a7b0430d8cdb78070b4c55a";
	}
	else if (AES_CLASS::Nk == 6) {
		// AES192
		key = "000102030405060708090a0b0c0d0e0f1011121314151617";
		plain = "00112233445566778899aabbccddeeff";
		cipher = "dda97ca4864cdfe06eaf70a0ec0d7191";
	}
	else if (AES_CLASS::Nk == 8) {
		// AES256
		key = "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f";
		plain = "00112233445566778899aabbccddeeff";
		cipher = "8ea2b7ca516745bfeafc49904b496089";
	}
	else {
		cout << "something wrong\n";
		return;
	}

	auto arrPlain = ByteArrayFromString(plain);
	auto arrKey = ByteArrayFromString(key);
	auto arrCipher = ByteArrayFromString(cipher);

	if (arrPlain.size() == 4 * AES_CLASS::Nb &&
		arrKey.size() == 4 * AES_CLASS::Nk &&
		arrCipher.size() == 4 * AES_CLASS::Nb)
	{
		// OK
	}
	else {
		cout << "something wrong\n";
		return;
	}

	{
		AES_CLASS aesEnc(arrKey.data(), false);

		vector<BYTE> arrTestCipher(4 * AES_CLASS::Nb);
		aesEnc.Cipher(arrPlain.data(), arrTestCipher.data());

		if (memcmp(arrTestCipher.data(), arrCipher.data(), 4 * AES_CLASS::Nb) != 0)
		{
			cout << "Cipher error\n";
			return;
		}
	}

	{
		AES_CLASS aesDec(arrKey.data(), true);

		vector<BYTE> arrTestInvCipher(4 * AES_CLASS::Nb);
		aesDec.EqInvCipher(arrCipher.data(), arrTestInvCipher.data());

		if (memcmp(arrTestInvCipher.data(), arrPlain.data(), 4 * AES_CLASS::Nb) != 0)
		{
			cout << "EqInvCipher error\n";
			return;
		}
	}

	cout << "OK\n";
}


int main()
{
	// CPUID`FbN
	{
		bool cpuok = false;

		int cpuinfo[4];
		__cpuid(cpuinfo, 0x01);
		bitset<32> cpuid_01_ecx(cpuinfo[2]);
		bitset<32> cpuid_01_edx(cpuinfo[3]);

		// edxbit25=SSEAbit26=SSE2Ă邱
		// ɂintel̃}jAɏĂ̂őfɏ]
		if (cpuid_01_edx[25] && cpuid_01_edx[26]) {
			// ecxbit25=AESNIĂ邱
			if (cpuid_01_ecx[25]) {

				// SSE4.1gĂ̂Ń`FbN
				if (cpuid_01_ecx[19]) {
					cpuok = true;
				}
			}
		}

		if (!cpuok) {
			cout << "cannot run on this CPU\n";
			return 1;
		}
	}


	cout << "Testing AES_SIMD...";
	test<AES_SIMD>();
	
	cout << "Testing AES128_SIMD...";
	test<AES128_SIMD>();

	cout << "Testing AES192_SIMD...";
	test<AES192_SIMD>();

	cout << "Testing AES256_SIMD...";
	test<AES256_SIMD>();

	return 0;
}

